/* @(#)autopart.c	1.10 98/06/01 Copyright 1995 J. Schilling */
#ifndef lint
static	char sccsid[] =
	"@(#)autopart.c	1.10 98/06/01 Copyright 1995 J. Schilling";
#endif
/*
 *	Automatic genation of partition tables
 *
 *	Copyright (c) 1995 J. Schilling
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/param.h>
#include <standard.h>
#include <string.h>
#include <dsklabel.h>

#include "fmt.h"
#include "scsireg.h"
#include "scsicmds.h"


extern	struct scsi_capacity cap;
struct	scsi_inquiry inq;

EXPORT	BOOL	get_part_defaults __PR((struct disk *, struct dk_label *));

#define	REST	(-1)			/* take the rest of the disk */
#define	INF	(~(unsigned long)0)	/* a very big disk */

struct ptab {
	int	part[NDKMAP];
};

struct default_parts {
	unsigned long	min_cap;
	unsigned long	max_cap;
	struct ptab	parts;
} default_parts[] = {
	{ 0,	64,	{{  0,   0, 0, 0, 0, 0, REST, 0 }} },
	{ 64,	180,	{{ 16,  16, 0, 0, 0, 0, REST, 0 }} },
#ifdef	SVR4
	{ 180,	380,	{{ 24,  32, 0, 0, 0, 0, REST, 0 }} },
	{ 380,	680,	{{ 24,  64, 0, 0, 0, 0, REST, 0 }} },
	{ 680,	1024,	{{ 24, 128, 0, 0, 0, 0, REST, 0 }} },
	{ 1024,	2048,	{{ 32, 128, 0, 0, 0, 0, REST, 0 }} },
	{ 2048,	INF,	{{ 64, 128, 0, 0, 0, 0, REST, 0 }} },
#else
	{ 180,	380,	{{ 16,  32, 0, 0, 0, 0, REST, 0 }} },
	{ 380,	680,	{{ 16,  64, 0, 0, 0, 0, REST, 0 }} },
	{ 680,	1024,	{{ 15, 128, 0, 0, 0, 0, REST, 0 }} },
	{ 1024,	2048,	{{ 24, 128, 0, 0, 0, 0, REST, 0 }} },
	{ 2048,	INF,	{{ 32, 128, 0, 0, 0, 0, REST, 0 }} },
#endif
};

#define	NPART_ENTS	(sizeof(default_parts) / sizeof(struct default_parts))

EXPORT
BOOL get_part_defaults(dp, lp)
	struct disk	*dp;
	register struct dk_label *lp;
{
	int	i;
	double	dcap;
	double	dmax;
	double	dmin;
	int	scyl;
	int	ncyl;
	long	l;
	struct ptab *pt;
	char	*p;

	if (cap.c_baddr <= 0) {
		if (read_capacity() < 0 && 
			dp->secsize > 0 && dp->capacity > 0) {
			cap.c_bsize = dp->secsize;
			cap.c_baddr = dp->capacity - 1;
		} else {
			return (FALSE);
		}
	}
	dcap = (cap.c_baddr+1.0) * (cap.c_bsize*1.0) / (1.024*1.024);
	for (i=0; i < NPART_ENTS; i++) {
		dmin = default_parts[i].min_cap * 1024.0 * 1024.0;
		dmax = default_parts[i].max_cap * 1024.0 * 1024.0;
		if (dcap >= dmin && dcap < dmax)
			break;
	}
	if (i == NPART_ENTS) {
		errmsgno(EX_BAD,
			"No autopart range found (this should not happen).\n");
		return (FALSE);
	}
	pt = &default_parts[i].parts;

	if (default_parts[i].max_cap == INF) {
 		printf("Using autopart range %lu MB - Infinity\n",
			default_parts[i].min_cap);
	} else {
		printf("Using autopart range %lu MB - %lu MB\n",
			default_parts[i].min_cap, default_parts[i].max_cap);
	}

	if (dp->lncyl < 0)
		dp->lncyl = get_default_lncyl(dp);

	scyl = 0;
	for (i=0; i < NDKMAP; i++) {
		if (pt->part[i]) {
			lp->dkl_map[i].dkl_cylno = scyl;
			l = pt->part[i];

			if (l == INF) {
				ncyl = dp->lncyl - scyl;
			} else {
				l *= 2048;
				ncyl = l / (dp->lhead * dp->lspt);

				if (l % (dp->lhead * dp->lspt))
					ncyl++;
			}
			l = ncyl * (dp->lhead * dp->lspt);
			lp->dkl_map[i].dkl_nblk = l;
			scyl += ncyl;
		}
	}

	/*
	 * Create ascilabel from vendor_info & prod_ident,
	 * skip multiple blanks.
	 */
	strncpy(lp->dkl_asciilabel, inq.vendor_info, sizeof(inq.vendor_info));
	p = &lp->dkl_asciilabel[sizeof(inq.vendor_info)-1];
	while (*p == ' ' && p > lp->dkl_asciilabel)
		p--;
	if (p > lp->dkl_asciilabel)
		p += 2;

	strncpy(p, inq.prod_ident, sizeof(inq.prod_ident));
	p[sizeof(inq.prod_ident)] = '\0';

	if ((p = strchr(p, ' ')) != NULL)
		*p = '\0';

	set_default_vtmap(lp);

	lp->dkl_magic = DKL_MAGIC;
	lp->dkl_cksum = do_cksum(lp);
	setlabel_from_val(dp, lp);
	return (TRUE);
}
